// bigint.h
// Little-Endian Big-Integer Class

#include <iostream.h>
#include <stdio.h>
#include "tuning.h"

#define BASE 10000
#define HIVALUES 999999999

#define IMAX(x,y) (((x) > (y)) ? (x) : (y))
#define IMIN(x,y) (((x) < (y)) ? (x) : (y))

extern "C" int BigMul(short w[], short u[], short B[], int n, int m);

////////////////////////////////////////////////////////////////////////////////
// BigData Class - Signed Arrays of Digits and Factors
////////////////////////////////////////////////////////////////////////////////

class BigData
{
    private:

        int count;
        int isign;
        int len;
        short *data;
        friend class BigInt;

    public:
        
		BigData()
        {
            count = 1;
            isign = +1;
            len = 0;
            data = new short[1];
        }
        
		BigData(int n)
        {
			int j;
            short temp[3];
            count = 1;
            isign = +1;
            len = 0;
            while (n > 0) {
                temp[len++] = (short) (n % 10000);
                n /= BASE;
            }
            data = new short[len];
            for (j=0; j<len; j++) data[j] = temp[j];
        }
        
		BigData(int n, int prec)
        {
			int j;
            short temp[3];
            count = 1;
            isign = +1;
            len = 0;
            while (n > 0) {
                temp[len++] = (short) (n % BASE);
                n /= BASE;
            }
            data = new short[prec];
            for (j=0; j<len; j++) data[j] = temp[j];
        }
        
		~BigData()
        {
            delete data;
        }
};

////////////////////////////////////////////////////////////////////////////////
// BigInt Class - only a Pointer to a BigData Object
////////////////////////////////////////////////////////////////////////////////

class BigInt
{
    private:
        
		BigData *p;

        BigInt add (BigInt);
        BigInt sub (BigInt);
    
	public:
        
		BigInt()
        {
            p = new BigData();
        }
        
		BigInt(int n)
        {
            p = new BigData(n);
        }
        
		BigInt(int n, int prec)
        {
            p = new BigData(n, prec);
        }
        
		BigInt(BigInt const &B)
        {
            p = B.p;
            (p->count)++;
        }
        
		~BigInt()
        {
            if (p->count == 1)
                delete p;
            else
                (p->count)--;
        }
        
		void operator = (BigInt const &B)
        {
            if (p->count == 1)
                delete p;
            else
                (p->count)--;
            p = B.p;
            (p->count)++;
        }
        
		void operator = (int n)
        {
            p = new BigData(n);
        }
        void sign(int s)
        {
            p->isign = s;
        }
        int get_size()
        {
            return p->len;
        }
        
        int copy(short *v, int n);
        BigInt operator + (BigInt);
        BigInt operator - (BigInt);
        BigInt operator * (BigInt);
#if DEBUG
		void show();
#endif
};

////////////////////////////////////////////////////////////////////////////////
// Display a Big Integer on the Console
////////////////////////////////////////////////////////////////////////////////

#if DEBUG
void BigInt::show()
{
	int j;
    char temp[5];
    if (p->isign == +1) cout << "+"; else cout << "-";
    for (j = p->len - 1; j >= 0; j--) {
        sprintf(temp,"%04d",p->data[j]);
        cout << temp;
    }
    cout << endl;
}
#endif

////////////////////////////////////////////////////////////////////////////////
// Copy a Big Integer to a Big-Endian 16-bit Array
////////////////////////////////////////////////////////////////////////////////

int BigInt::copy(short *v, int n)
{
	int j;
    for (j = 0; (j < p->len) && (j < n); j++)
        v[j] = p->data[p->len-1-j];
    return p->len;
}

////////////////////////////////////////////////////////////////////////////////
// Add two Unsigned Big Integers (private)
////////////////////////////////////////////////////////////////////////////////

BigInt BigInt::add (BigInt B)
{
	int j;
    int maxlen = (p->len > B.p->len) ? p->len : B.p->len;
    BigInt temp(0,maxlen+1);
    int carry = 0;
    for (j = 0; j < maxlen; j++) {
        int d1 = (j > p->len - 1) ? 0 : p->data[j];
        int d2 = (j > B.p->len - 1) ? 0 : B.p->data[j];
        int digitsum = d1 + d2 + carry;
        if (digitsum >= BASE) { 
            digitsum -= BASE;
            carry = 1;
        } else {
            carry = 0;
        }
        temp.p->data[j] = digitsum;
    }
    if (carry == 1)
         temp.p->data[j++] = 1;
    temp.p->len = j;
    return temp;
}

////////////////////////////////////////////////////////////////////////////////
// Subtract two Unsigned Big Integers (private)
////////////////////////////////////////////////////////////////////////////////

BigInt BigInt::sub (BigInt B)
{
    int maxlen = (p->len > B.p->len) ? p->len : B.p->len;
    BigInt temp(0,maxlen);
    int carry=0, j, k=-1, dir=+1, digitsum;
    for (j = maxlen - 1; j >=0; j--) {
        int d1 = (j > p->len - 1) ? 0 : p->data[j];
        int d2 = (j > B.p->len - 1) ? 0 : B.p->data[j];
        if (d1 > d2) { dir = +1; break; }
        if (d1 < d2) { dir = -1; break; }
    }
    for (j = 0; j < maxlen; j++) {
        int d1 = (j > p->len - 1) ? 0 : p->data[j];
        int d2 = (j > B.p->len - 1) ? 0 : B.p->data[j];
        if (dir == +1)
            digitsum = d1 - d2 - carry;
        else
            digitsum = d2 - d1 - carry;
        if (digitsum < 0) { 
            digitsum += BASE;
            carry = 1;
        } else {
            carry = 0;
        }
        temp.p->data[j] = digitsum;
        if (temp.p->data[j] != 0) k = j;
    }
    temp.p->len = ++k;
    temp.p->isign = dir;
    return temp;
}

////////////////////////////////////////////////////////////////////////////////
// Operator Overload for Plus Sign
////////////////////////////////////////////////////////////////////////////////

BigInt BigInt::operator + (BigInt B)
{
    int maxlen = (p->len > B.p->len) ? p->len : B.p->len;
    BigInt temp(0,maxlen+1);
    if (p->isign == B.p->isign) {
        temp = add(B);
        temp.p->isign = p->isign;
    } else {
		temp = sub(B);
		if (temp.p->isign == +1)
			temp.p->isign = p->isign;
		else
			temp.p->isign = -(p->isign);
	}
    return temp;
}

////////////////////////////////////////////////////////////////////////////////
// Operator Overload for Minus Sign
////////////////////////////////////////////////////////////////////////////////

BigInt BigInt::operator - (BigInt B)
{
    int maxlen = (p->len > B.p->len) ? p->len : B.p->len;
    BigInt temp(0,maxlen+1);
    if (p->isign != B.p->isign) {
        temp = add(B);
        temp.p->isign = p->isign;
    } else {
		temp = sub(B);
		if (temp.p->isign == +1)
			temp.p->isign = p->isign;
		else
			temp.p->isign = -(p->isign);
	}
    return temp;
}

////////////////////////////////////////////////////////////////////////////////
// Operator Overload for Multiply Sign
////////////////////////////////////////////////////////////////////////////////

BigInt BigInt::operator * (BigInt B)
{
    int i, j, k, maxlen, q, r, s;
    maxlen = p->len + B.p->len;
    BigInt temp(0,maxlen);
    if (!p->len || !B.p->len)
        return temp;
    temp.p->len = maxlen;
    if (IMIN(p->len, B.p->len) >= MIN_FFTMUL) {
        BigMul(temp.p->data, p->data, B.p->data, p->len,  B.p->len);
    } else {
        for (k = 0; k < maxlen; k++)
            temp.p->data[k] = 0;
        for (i = 0; i < p->len; i++) {
            q = 0;
            for (j = 0; j < B.p->len; j++) {
                k = i + j;
                s = (int) p->data[i] * B.p->data[j] + temp.p->data[k] + q;
                q = s / BASE;
                r = s - BASE * q;
                temp.p->data[k] = (short) r;
            }
            if (q) temp.p->data[++k] = (short) q;
        }
    }
    temp.p->isign = p->isign * B.p->isign;
    if (temp.p->data[maxlen-1] == 0)
        temp.p->len = maxlen - 1;
    return temp;
}

