// bigmul.c
// Little-Endian Big-Integer Multiply
// using David Baileys Non-Power of 2 Recursive Method

#include "mpmath.h"
#include "tuning.h"

#define IMAX(x,y) (((x) > (y)) ? (x) : (y))
#define IMIN(x,y) (((x) < (y)) ? (x) : (y))

double* LinearConvolution(short num1[], short num2[], int len1, int len2);
double* SmallLinearConvolution(short num1[], short num2[], int len1, int len2);
void    dfour1(double data[], unsigned long nn, int isign);
void    drealft(double data[], unsigned long n, int isign);

void BigMul(short prod[], short num1[], short num2[], int len1, int len2)
/* multiple precision multiply, using FFT methods */
{
	#define RX 10000.0
    int j,maxlen,nn=1,len3=len1+len2;
    double cy,t,*b;

    maxlen=IMAX(len2,len1);
	while (nn < maxlen) nn <<= 1;
	nn <<= 1;
	b=LinearConvolution(num1,num2,len1,len2);
    cy=0.0;
    for (j=nn-1;j>=0;j--) {
        t=b[j]+cy+0.5;
        cy=(unsigned long) (t/RX);
        b[j]=t-cy*RX;
    }
    if (cy >= RX) error("cannot happen in fftmul");

    prod[len3-1]=(short) cy;
    for (j=1;j<len3;j++)
        prod[len3-j-1]=(short) b[j-1];

	free_dvector(b);
}

double* LinearConvolution(short num1[], short num2[], int len1, int len2)
{
    int i,j,k,maxlen,nn=1,p=0,m=0,recurse=0;
    double t,*a,*b,*b2;

    maxlen=IMAX(len2,len1);
	while (nn < maxlen) nn <<= 1;
	nn <<= 1;

	if (maxlen >= 2) {
		p = 1; while (p < maxlen) p <<= 1;
		if (p > maxlen) p >>= 1;
		m = maxlen - p;
		recurse = (m > 0 && m <= (p>>1));
	}

	a=dvector(nn);
	b=dvector(nn);

    for (j=0;j<len1;j++)
        a[j]=(double)num1[len1-1-j];
    for (j=len1;j<nn;j++) a[j]=0.0;
    for (j=0;j<len2;j++)
        b[j]=(double)num2[len2-1-j];
    for (j=len2;j<nn;j++) b[j]=0.0;

	if (recurse) nn >>= 1;

    drealft(a,nn,1);
	if (num1==num2 && len1==len2) {
	    for (j=0;j<nn;j++)
			b[j]=a[j];
	} else {
		drealft(b,nn,1);
	}

    b[0] *= a[0];
    b[1] *= a[1];
    for (j=2;j<nn;j+=2) {
        b[j]=(t=b[j])*a[j]-b[j+1]*a[j+1];
        b[j+1]=t*a[j+1]+b[j+1]*a[j];
    }
    drealft(b,nn,-1);
    for (j=nn-1;j>=0;j--)
        b[j]=b[j]/(nn>>1);
	free_dvector(a);

	if (recurse) {
		k=p-m+1;
		if (IMIN(len1,len2)-k>0) {
			if (m<MIN_FFTCONV)
				b2=SmallLinearConvolution(num1,num2,len1-k,len2-k);
			else
				b2=LinearConvolution(num1,num2,len1-k,len2-k);
			j=2*m-2;
			for (i=0;i<2*m-1;i++,j++) {
				b[i]-=b2[j];
				b[2*p+i]=b2[j];
			}
			free_dvector(b2);
		}
	}

	return b;
}

double* SmallLinearConvolution(short num1[], short num2[], int len1, int len2)
{
    int i,j,k,k1,k2,n,nn;
    double *b;

    n=IMAX(len2,len1);
	nn=n<<1;
	k1=n-1;
	k2=2*n-2;
	b=dvector(nn);
	for (i=0;i<nn;i++)
		b[i]=0.0;
	for (i=0;i<len1;i++) {
		for (j=0;j<len2;j++) {
			k=i+j;
			if (k>=k1 && k<=k2) {
				b[k]=b[k]+num1[len1-1-i]*num2[len2-1-j];
			}
		}
	}
	return b;
}
